<?php

namespace Drupal\donorperfect;

use Drupal\Component\Utility\NestedArray;

class DPQuery {

  /**
   * The DonorPerfect XML API endpoint URL.
   */
  const DONORPERFECT_API_URL = 'https://www.donorperfect.net/prod/xmlrequest.asp';

  /**
   * List the valid types of queries that can be run by DPQuery.
   *
   * @var array
   */
  protected $valid_types = [
    'select',
    'insert',
    'update',
    'pass',
    'retrieveAllTableData',
  ];

  /**
   * List the DonorPerfect predefined procedures that can be run by DPQuery.
   *
   * @var array
   */
  protected $predefined_procs = [
    'donorsearch',
    'gifts',
    'savedonor',
    'savegift',
    'saveotherinfo',
    'savecontact',
    'delflags',
    'saveflag',
    'saveudf',
  ];

  /**
   * List the valid types of queries that can be run by DPQuery.
   *
   * @var array
   */
  protected $is_proc = FALSE;

  /**
   * Array of fields required for the predefined procedure being used.
   *
   * @var array
   */
  protected $proc_fields = [];

  /**
   * List the tables that are accessible by the XML API in DonorPerfect for queries.
   *
   * @var array
   */
  protected $avail_tables = [
    'dp',
    'dpudf',
    'dpgift',
    'dpgiftudf',
    'dpaddress',
    'dpaddressmailings',
    'dplink',
    'dpotherinfo',
    'dpotherinfoudf',
    'dppaymentmethod',
    'dpcontact',
    'dpcontactudf',
    'dpusermultivalues',
    'dpcodes',
    'dpflags',
    'dpbatch',
  ];

  /**
   * Array of fields accessible by the XML API query, keyed by table name.
   *
   * @var array
   */
  protected $avail_fields = [];

  /**
   * Authentication string to be used for calls to the XML API.
   *
   * @var string
   */
  protected $auth_string;

  /**
   * Set the type of query: select, insert, update, delete, or
   *   pass (for a pass-thru sql statement).
   *
   * @var string
   */
  protected $type = '';

  /**
   * DonorPerfect table to be used as the base table for the query.
   *
   * @var string
   */
  protected $base_table = '';

  /**
   * DonorPerfect field (from $base_table) to be used as the base field for
   *   the query.
   *
   * @var string
   */
  protected $base_field = '';

  /**
   * Array of fields to be retrieved by a select query.
   * This should be a 2-dimensional array of tables=>fields.
   *
   * @var array
   */
  protected $tables_fields = [];

  /**
   * Array of SQL conditions for the query.
   *
   * @var array
   */
  protected $conditions = [];

  /**
   * Array of ORDER BY clauses for the query.
   *
   * @var array
   */
  protected $order_by = [];

  /**
   * ORDER BY clauses imploded to a string.
   *
   * @var string
   */
  protected $order_by_string = '';

  /**
   * Array of fields and values to be updated by an update query.
   *
   * @var array
   */
  protected $update_fields = [];

  /**
   * Complied SQL statement.
   *
   * @var string
   */
  protected $sql = '';

  /**
   * Value of the 'action' parameter in the API call,
   *  based on which predefined procedure/query is being run.
   *
   * @var string
   */
  protected $dp_action = '';

  /**
   * Array of parameters passed to the query object.
   *
   * @var array
   */
  protected $params = [];

  /**
   * Query object parameters compiled to a string.
   *
   * @var string
   */
  protected $params_string = '';

  /**
   * Final compiled API call string.
   *
   * @var string
   */
  protected $api_call = '';

  /**
   * Type of SQL join to be used in the query.
   *
   * @var string
   */
  protected $join_type = ' INNER JOIN ';

  /**
   * Maximum number of records to be returned by a select query.
   *
   * @var int|null
   */
  protected $limit = NULL;

  /**
   * Offset for paged query results.
   *
   * @var int|null
   */
  protected $offset = NULL;

  /**
   * Array of options to pass to the http client.
   *
   * @var array
   */
  protected $http_request_options = ['timeout' => 30];

  /**
   * Boolean indicating whether or not the query should return a count
   *  rather than specific fields.
   *
   * @var bool
   */
  public $get_count = FALSE;

  /**
   * HTTP result code for the API call (200 for success).
   *
   * @var string|int
   */
  public $result_code = '';

  /**
   * Array of query results.
   *
   * @var array
   */
  public $result = [];

  /**
   * Name of the currently logged-in user.
   *
   * @var string
   */
  public $user_name = '';

  /**
   * Instance constructor.
   *
   * @param string $type
   *   The type of query object to create. Can be any of $valid_types or
   *   $predefined_procs defined above.
   * @param mixed $optional
   *   Optional additional information for the query object, depending on
   *   the $type of query object being created.
   */
  public function __construct(string $type = '', mixed $optional = null) {
    $this->valid_types = array_merge($this->valid_types, $this->predefined_procs);

    if (in_array($type, $this->valid_types)) {
      $this->type = $type;
      $this->is_proc = in_array($type, $this->predefined_procs);
      if ($this->is_proc) {
        $this->proc_fields = $this->getProcFields();
      }
    } else {
      $this->type = '';
    }

    // Prepare authentication string.
    $this->auth_string = '';
    $api_key = \Drupal::config('donorperfect.settings')->get('api.key');
    if (!empty($api_key)) {
      $this->auth_string = "apikey={$api_key}";
    } else {
      $dp_user = \Drupal::config('donorperfect.settings')->get('api.user');
      $dp_pw = \Drupal::config('donorperfect.settings')->get('api.pw');
      if (!empty($dp_user) && !empty($dp_pw)) {
        $this->auth_string = "login={$dp_user}&pass={$dp_pw}";
      }
    }

    $this->avail_fields = \Drupal::config('donorperfect.cache')->get('dptables') ?? [];

    // For SELECT, INSERT, and UPDATE queries, $optional can be either
    // the 'base_table' or 'base_table.base_field'.
    if ((in_array($this->type, ['select','insert','update'])) && !empty($optional) && is_string($optional)) {
      $thisDot = strpos($optional, '.');
      if ($thisDot !== FALSE) {
        $thisTable = trim(substr($optional, 0, $thisDot));
        $thisField = trim(substr($optional, $thisDot + 1, strlen($optional) - 1));
        if (array_key_exists($thisTable, $this->avail_fields) && array_key_exists($thisField, $this->avail_fields[$thisTable])) {
          $this->base_table = $thisTable;
          $this->base_field = $thisField;
        }
      } else {
        if (in_array($optional, $this->avail_tables)) {
          $this->base_table = $optional;
          $this->base_field = $this->getBaseField($optional);
        }
      }
      if (($this->type === 'select') && !empty($this->base_table) && !empty($this->base_field)) {
        $this->addField("{$this->base_table}.{$this->base_field}");
      }
    }

    // For passthrough queries, $optional is the SQL statement to be executed.
    if (($this->type == 'pass') && $optional) {
      $this->sql = $optional;
    }

    if (($this->type == 'retrieveAllTableData') && $optional) {
      //We use this procedure to retrieve all the data from a specified table.
      $this->base_table = $optional;
    }

    // Set user_name.
    $this->user_name = \Drupal::currentUser()->getDisplayName();
  }

  /**
   * For SELECT queries, this function adds a field to the list of fields to retrieve.
   * For UPDATE queries, adds a field and a value to be updated by the query.
   * Returns $this to enable daisy-chaining.
   *
   * @param string $field
   * @param mixed $value
   * @param bool $add_quotes
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function addField(string $field, mixed $value = null, bool $add_quotes = FALSE) {
    if (!empty($field) && is_string($field)) $field = strtolower($field);
    if ($this->is_proc) {
      //If $field is one of the procedure parameters defined by DonorPerfect, set $add_quotes based on DP's definition.
      if (array_key_exists($field, $this->proc_fields)) {
        //If the proc_field is a date or character, then add quotes.
        if (in_array($this->proc_fields[$field], ['d','c'])) {
          $add_quotes = TRUE;
        } else {
          $add_quotes = FALSE;
        }
      }
      //Add field to $this->params.
      $this->params[$field] = $add_quotes ? $this->addSingleQuotes($value) : $value;
    } elseif ($this->type == 'select') {
      if (!empty($field)) {
        $thisDot = strpos($field, '.');
        if ($thisDot !== FALSE) {
          $thisTable = trim(substr($field, 0, $thisDot));
          $thisField = trim(substr($field, $thisDot + 1, strlen($field) - 1));
        } else {
          $thisTable = $this->base_table;
          $thisField = trim($field);
        }
        if (array_key_exists($thisTable, $this->avail_fields)) {
          if (($thisField === '*') || array_key_exists($thisField, $this->avail_fields[$thisTable])) {
            // Check if this is a multi-value field.
            if (empty(NestedArray::getValue($this->avail_fields, [$thisTable, $thisField, 'multi']))) {
              if ((array_key_exists($thisTable, $this->tables_fields) && is_array($this->tables_fields[$thisTable]) && !in_array($thisField, $this->tables_fields[$thisTable]) && !in_array('*', $this->tables_fields[$thisTable])) ||
                  !array_key_exists($thisTable, $this->tables_fields)) {
                if ($thisField === '*') {
                  // If requesting * for this table, remove all individually-listed fields.
                  $this->tables_fields[$thisTable] = ['*'];
                } else {
                  $this->tables_fields[$thisTable][] = $thisField;
                }
              }
            } else {
              // This is a multi-value field.
              if ($thisField === 'flag') {
                $thisTable = 'dpflags';
              } else {
                $thisTable = 'dpusermultivalues';
              }
              if ((array_key_exists($thisTable, $this->tables_fields) && is_array($this->tables_fields[$thisTable]) && !in_array($thisField, $this->tables_fields[$thisTable])) ||
                  !array_key_exists($thisTable, $this->tables_fields)) {
                $this->tables_fields[$thisTable][] = $thisField;
              }
            }
          } else {
            // Check if this is a field on the related UDF table.
            $udfTable = "{$thisTable}udf";
            if (array_key_exists($udfTable, $this->avail_fields)) {
              if (array_key_exists($thisField, $this->avail_fields[$udfTable])) {
                // Check if this is a multi-value field.
                if (empty(NestedArray::getValue($this->avail_fields, [$udfTable, $thisField, 'multi']))) {
                  if ((array_key_exists($udfTable, $this->tables_fields) && is_array($this->tables_fields[$udfTable]) && !in_array($thisField, $this->tables_fields[$udfTable])) ||
                      !array_key_exists($udfTable, $this->tables_fields)) {
                    $this->tables_fields[$udfTable][] = $thisField;
                  }
                } else {
                  // This is a multi-value field.
                  $udfTable = 'dpusermultivalues';
                  if ((array_key_exists($udfTable, $this->tables_fields) && is_array($this->tables_fields[$udfTable]) && !in_array($thisField, $this->tables_fields[$udfTable])) ||
                      !array_key_exists($udfTable, $this->tables_fields)) {
                    $this->tables_fields[$udfTable][] = $thisField;
                  }
                }
              }
            }
          }
        }
      }
    } elseif (in_array($this->type, ['insert','update'])) {
      if ($field && !is_null($value)) {
        if (array_key_exists($this->base_table, $this->avail_fields)) {
          if (array_key_exists($field, $this->avail_fields[$this->base_table])) {
            $this->update_fields[$field] = $add_quotes ? $this->addSingleQuotes($value) : $value;
          }
        }
      }
    }
    return $this;    
  }

  /**
   * Returns boolean indicating whether or not the specified field has been
   *   added to the query.
   *
   * @param string $field
   *
   * @return bool
   */
  public function hasField(string $field) {
    $output = FALSE;
    if (!empty($field) && is_string($field)) {
      $field = strtolower($field);
      if ($this->is_proc) {
        $output = array_key_exists($field, $this->params);
      } else {
        if (in_array($this->type, ['insert','update'])) {
          $output = array_key_exists($field, $this->update_fields);
        } elseif ($this->type === 'select') {
          $thisDot = strpos($field, '.');
          if ($thisDot !== FALSE) {
            $thisTable = trim(substr($field, 0, $thisDot));
            $thisField = trim(substr($field, $thisDot + 1, strlen($field) - 1));
          } else {
            $thisTable = $this->base_table;
            $thisField = trim($field);
          }
          $exists = FALSE;
          NestedArray::getValue($this->tables_fields, [$thisTable, $thisField], $exists);
          $output = $exists;
        }
      }
    }
    return $output;
  }

  /**
   * Removes a field from the query.
   *
   * @param string $field
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function removeField(string $field) {
    if (!empty($field) && is_string($field)) {
      $field = strtolower($field);
      if ($this->is_proc) {
        if (array_key_exists($field, $this->params)) {
          unset($this->params[$field]);
        }
      } elseif ($this->type === 'select') {
        $thisDot = strpos($field, '.');
        if ($thisDot !== FALSE) {
          $thisTable = trim(substr($field, 0, $thisDot));
          $thisField = trim(substr($field, $thisDot + 1, strlen($field) - 1));
        } else {
          $thisTable = $this->base_table;
          $thisField = trim($field);
        }
        if (!empty($this->tables_fields[$thisTable])) {
          $key = array_search($thisField, $this->tables_fields[$thisTable]);
          if ($key !== FALSE) {
            unset($this->tables_fields[$thisTable][$key]);
          }
        }
      } elseif (in_array($this->type, ['insert','update'])) {
        if (array_key_exists($field, $this->update_fields)) {
          unset($this->update_fields[$field]);
        }
      }
    }
    return $this;
  }

 /**
  * For SELECT and UPDATE queries, this function adds a condition to the list of conditions to narrow results.
  * Returns the object to enable daisy-chaining.
  *
  * @param string $condition
  *
  * @return \Drupal\donorperfect\DPQuery $this
  */
  public function addCondition(string $condition) {
    if (in_array($this->type, ['select','update'])) {
      if (!empty($condition) && is_string($condition)) {
        $this->conditions[] = $condition;
      }
    }
    return $this;
  }

 /**
  * For SELECT queries, this function adds a field to order the results by.
  * Returns the object to enable daisy-chaining.
  *
  * @param string $order
  *
  * @return \Drupal\donorperfect\DPQuery $this
  */
  public function addOrder(string $order) {
    if ($this->type == 'select') {
      if (!empty($order) && is_string($order)) {
        $this->order_by[] = $order;
      }
    }
    return $this;
  }

 /**
  * For predefined procedures, this function adds a parameter to the query string.
  * Returns the object to enable daisy-chaining.
  * 
  * This is now just an alias for addField, and remains here to support legacy code.
  *
  * @param string $field
  * @param mixed $value
  * @param bool $add_quotes
  *
  * @return \Drupal\donorperfect\DPQuery $this
  */
  public function addParam(string $field, mixed $value, bool $add_quotes=FALSE) {
    return $this->addField($field, $value, $add_quotes);
  }

  /**
   * Set the type of join between the tables in the query.
   *
   * @param string $type
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setJoinType(string $type = 'inner') {
    switch ($type) {
      case 'left':
        $this->join_type = ' LEFT JOIN ';
        break;
      case 'inner':
      default:
        $this->join_type = ' INNER JOIN ';
        break;
    }
    return $this;
  }

  /**
   * Set $this->sql for passthrough queries.
   *
   * @param string $sql
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setSql(string $sql) {
    if ($this->type === 'pass') {
      $this->sql = $sql;
    }
    return $this;
  }

  /**
   * Set the max time (in number of seconds) allowed to execute the api call.
   *
   * @param float $seconds
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setTimeout(float $seconds) {
    if (is_numeric($seconds)) $this->http_request_options['timeout'] = floatval($seconds);
    return $this;
  }

  /**
   * Set an option to be used by the HTTP client when making the call to the
   *   DonorPerfect API. See \GuzzleHttp\RequestOptions for available options.
   *
   * @param string $key
   *   The name of the option to set.
   * @param mixed $value
   *   The value to be set for the option.
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setHttpOption(string $key, mixed $value) {
    $this->http_request_options[$key] = $value;
    return $this;
  }

  /**
   * Set the max number of records to be returned by the query.
   *
   * @param int $new_limit
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setLimit(int $new_limit) {
    $this->limit = $new_limit;
    return $this;
  }

  /**
   * Return the configured max number of records to be returned by the query.
   *
   * @return int|null
   */
  public function getLimit() {
    return $this->limit;
  }

  /**
   * Set the offset for paginated query results.
   *
   * @param int $new_offset
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function setOffset(int $new_offset) {
    $this->offset = $new_offset;
    return $this;
  }

  /**
   * Return the configured offset for paginated query results.
   *
   * @return int
   */
  public function getOffset() {
    return $this->offset;
  }

  /**
   * Shortcut function to set both offset and limit in a single call.
   *
   * @param int $offset
   * @param int $limit
   *
   * @return \Drupal\donorperfect\DPQuery $this
   */
  public function queryRange($offset, $limit) {
    $this->offset = $offset;
    $this->limit = $limit;
    return $this;
  }

  /**
   * Return array of all fields available to be queried, keyed by
   *   table name.
   *
   * @return array
   */
  public function listFields() {
    return $this->avail_fields;
  }

  /**
   * Return the compiled ORDER BY clause.
   *
   * @return string
   */
  public function getOrder() {
    return $this->order_by_string;
  }

  /**
   * Compile query tables and fields into strings for use in the
   *   SQL statement.
   *
   * @return string[]
   */
  public function compileTablesFields() {
    $varFrom = $varFields = '';
    if (count($this->tables_fields) > 0) {
      $x = 0;
      $joinedTables = [];
      $baseTable = $baseTableMain = '';
      foreach ($this->tables_fields as $table => $fields) {
        // Reset $baseTable in case it had to be changed in a previous iteration.
        $baseTable = $baseTableMain;
        if (in_array($table, $this->avail_tables)) {
          //Prepare FROM portion of select statement.
          if ($x == 0) {
            $varFrom = $table;
            $baseTable = $baseTableMain = $table;
            $joinedTables[] = $table;
          } else {
            $joinHandled = FALSE;
            switch ($table) {
              case 'dpaddressmailings':
                if (!in_array('dpaddressmailings', $joinedTables)) {
                  $temp = "{$this->join_type}dpaddressmailings ON dpaddress.address_id=dpaddressmailings.address_id";
                  if (in_array('dpaddress', $joinedTables)) {
                    if ($x > 1) {
                      $varFrom = "({$varFrom})";
                    }
                    $varFrom = "({$varFrom}){$temp}";
                    $joinedTables[] = 'dpaddressmailings';
                  } else {
                    if ($x > 1) {
                      $varFrom = "({$varFrom})";
                    }
                    $varFrom = "({$varFrom}{$this->join_type}dpaddress ON {$baseTable}.donor_id=dpaddress.donor_id){$temp}";
                    $joinedTables[] = 'dpaddress';
                    $joinedTables[] = 'dpaddressmailings';
                  }
                }
                $joinHandled = TRUE;
                break;
              case 'dpusermultivalues':
                if (!in_array('dpusermultivalues', $joinedTables)) {
                  if (count($fields) === 1) {
                    // @todo Make it possible to handle more than one multi-value field in a single query.
                    $joinField = '';
                    switch ($baseTable) {
                      case 'dpgift':
                        $joinField = 'dpgift.gift_id';
                        break;
                      case 'dpcontact':
                        $joinField = 'dpcontact.contact_id';
                        break;
                      case 'dpotherinfo':
                        $joinField = 'dpotherinfo.other_id';
                        break;
                      case 'dp':
                        $joinField = 'dp.donor_id';
                        break;
                    }
                    if (!empty($joinField)) {
                      $field_name = strtoupper((string)reset($fields));
                      $temp = "{$this->join_type}dpusermultivalues ON {$joinField}=dpusermultivalues.matching_id AND dpusermultivalues.field_name='{$field_name}'";
                      if ($x > 1) {
                        $varFrom = "({$varFrom})";
                      }
                      $varFrom = "{$varFrom}$temp";
                      $joinedTables[] = 'dpusermultivalues';
                    } else {
                      // Don't add any fields from this table.
                      $fields = [];
                    }
                  } else {
                    // Don't add any fields from this table.
                    $fields = [];
                  }
                }
                $joinHandled = TRUE;
                break;
              case 'dpflags':
                if (!in_array('dpflags', $joinedTables)) {
                  $joinField = '';
                  switch ($baseTable) {
                    case 'dp':
                    case 'dpgift':
                    case 'dpcontact':
                    case 'dpotherinfo':
                      $joinField = "{$baseTable}.donor_id";
                      break;
                  }
                  if (!empty($joinField)) {
                    $temp = "{$this->join_type}dpflags ON {$joinField}=dpflags.donor_id";
                    if ($x > 1) {
                      $varFrom = "({$varFrom})";
                    }
                    $varFrom = "{$varFrom}$temp";
                    $joinedTables[] = 'dpflags';
                  } else {
                    // Don't add any fields from this table.
                    $fields = [];
                  }
                }
                $joinHandled = TRUE;
                break;
            }
            if (!$joinHandled && !in_array($table, $joinedTables)) {
              // If this is a UDF table, make sure the parent table is joined first.
              $matches = [];
              if (preg_match('/^(.+)udf$/', $table, $matches)) {
                if (!in_array($matches[1], $joinedTables)) {
                  if ($x > 1) {
                    $varFrom = "({$varFrom})";
                  }
                  $varFrom = "{$varFrom}{$this->join_type}{$matches[1]} ON {$baseTable}.donor_id={$matches[1]}.donor_id";
                  // Record that the parent table has been joined.
                  $joinedTables[] = $matches[1];
                  // Increment $x because another table was just added.
                  $x++;
                }
                // Temporarily set $baseTable to the parent table in order to
                // join to the parent table rather than $baseTableMain.
                $baseTable = $matches[1];
              }
              if ($x > 1) {
                $varFrom = "({$varFrom})";
              }
              $joinField = 'donor_id';
              switch ($baseTable) {
                case 'dpgift':
                  if ($table === 'dpgiftudf') $joinField = 'gift_id';
                  break;
                case 'dpcontact':
                  if ($table === 'dpcontactudf') $joinField = 'contact_id';
                  break;
                case 'dpotherinfo':
                  if ($table === 'dpotherinfoudf') $joinField = 'other_id';
                  break;
              }
              $varFrom = "{$varFrom}{$this->join_type}{$table} ON {$baseTable}.{$joinField}={$table}.{$joinField}";
              $joinedTables[] = $table;
              $joinHandled = TRUE;
            }
          }
          $x++;
          // Prepare FIELDS portion of select statement.
          foreach($fields as $field) {
            if ($table == 'dpaddress') {
              // For the dpaddress table, add field aliases so that the values don't overwrite
              // values from the dp table with same field names.
              $varFields .= "{$table}.{$field} sub_{$field},";
            } elseif ($table == 'dpusermultivalues') {
              if (strpos($varFields, $table) === FALSE) {
                $varFields .= "{$table}.matching_id, {$table}.code,";
              }
            } elseif ($table == 'dpflags') {
              if (strpos($varFields, $table) === FALSE) {
                $varFields .= "{$table}.donor_id flag_donor_id, {$table}.flag,";
              }
            } else {
              $varFields .= "{$table}.{$field},";
            }
          }
        }
      }
      //Remove the final comma from varFields.
      if (!empty($varFields)) {
        $varFields = substr($varFields, 0, strlen($varFields) - 1);
      }
    }
    return [$varFrom, $varFields];
  }

  /**
   * Compile the WHERE clause for the SQL statement.
   *
   * @return string
   */
  public function compileWhere() {
    $varWhere = '';
    $x = 0;
    foreach ($this->conditions as $condition) {
      if ($x === 0) {
        $varWhere = "({$condition})";
      } else {
        $varWhere .= " AND ({$condition})";
      }
      $x++;
    }
    return $varWhere;
  }

 /**
  * This function routes query execution to the appropriate function.
  * Returns the object to enable daisy-chaining.
  *
  * @return \Drupal\donorperfect\DPQuery $this
  */
  public function execute() {
    $thisFunc = "execute_{$this->type}";
    if (method_exists($this, $thisFunc)) {
      $this->$thisFunc();
    }
    return $this;
  }

  /**
   * Executes a SQL SELECT statement.
   */
  protected function execute_select() {
    //Compile tables and fields.
    list($varFrom, $varFields) = $this->compileTablesFields();
    if (!empty($varFrom) && !empty($varFields)) {
      //If this is a count query, replace the FIELDS portion of the select statement.
      if ($this->get_count) {
        if (!empty($this->base_table) && !empty($this->base_field)) {
          $varFields = "COUNT({$this->base_table}.{$this->base_field}) AS count";
        } else {
          foreach ($this->tables_fields as $table=>$fields) {
            $varFields = "COUNT({$table}.{$fields[0]}) AS count";
            break;
          }
        }
      }
      //Prepare WHERE portion of select statement.
      $varWhere = $this->compileWhere();
      //Prepare ORDER BY portion of select statement.
      $varOrder = '';
      if (!empty($this->order_by) && is_array($this->order_by)) {
        $varOrder = implode(',', $this->order_by);
      }
      $this->order_by_string = $varOrder;
      //Prepare final select statement.
      if ($this->get_count) {
        $this->sql = "SELECT {$varFields} FROM {$varFrom}";
        if (!empty($varWhere)) $this->sql .= " WHERE {$varWhere}";
        //Execute API call.
        $this->execute_select_api_call();
      } elseif (is_null($this->offset) || is_null($this->limit)) {
        //Get results in groups of 500 because of DP API's limit of returning a max of 500 records.
        if (empty($varOrder) && !empty($this->base_table) && !empty($this->base_field)) {
          $varOrder = "{$this->base_table}.{$this->base_field} ASC";
        }
        $x = -499;
        $y = 0;
        $result = [];
        $loop = TRUE;
        while ($loop) {
          $x += 500;
          $y += 500;
          $this->sql = "WITH mytable AS (SELECT {$varFields},ROW_NUMBER() OVER (ORDER BY {$varOrder}) AS RowNumber FROM {$varFrom} WHERE {$varWhere}) SELECT * FROM myTable WHERE (RowNumber BETWEEN {$x} AND {$y})";
          $this->execute_select_api_call();
          $result = array_merge($result, $this->result);
          if (count($this->result) < 500) {
            $this->result = $result;
            $loop = FALSE;
          }
        }
      } else {
        //Use offset and limit to get specific records.
        $x = $this->offset + 1;
        $y = $this->offset + $this->limit;
        if (empty($varOrder) && $this->base_table && $this->base_field) $varOrder = $this->base_table . '.' . $this->base_field . ' ASC';
        $varWhere = !empty($varWhere) ? ' WHERE ' . $varWhere : '';
        $this->sql = "WITH mytable AS (SELECT {$varFields},ROW_NUMBER() OVER (ORDER BY {$varOrder}) AS RowNumber FROM {$varFrom}{$varWhere}) SELECT * FROM myTable WHERE (RowNumber BETWEEN {$x} AND {$y})";
        //Execute API call.
        $this->execute_select_api_call();
      }
    }
  }

  /**
   * This intermediary function for execute_api_call caches the results of each SELECT query
   *  and, if the same SELECT query is passed again, returns the previous results without pinging the DP API again.
   */
  protected function execute_select_api_call() {
    $previous_select_calls = &drupal_static('dpquery_select_calls');
    if (!is_array($previous_select_calls)) $previous_select_calls = [];
    if (array_key_exists($this->sql, $previous_select_calls)) {
      $this->result = $previous_select_calls[$this->sql]['result'];
      $this->result_code = $previous_select_calls[$this->sql]['result_code'];
    } else {
      $this->execute_api_call();
      $previous_select_calls[$this->sql] = [
        'result' => $this->result,
        'result_code' => $this->result_code,
      ];
    }
  }

  /**
   * Executes a SQL INSERT statement.
   */
  protected function execute_insert() {
    if (($this->base_table) && (count($this->update_fields) > 0)) {
      //If not already included, add the currently logged-in website user and date as modified_by and modified_date.
      if (!array_key_exists('created_by',$this->update_fields) || !array_key_exists('created_date',$this->update_fields)) {
        $thisUser = 'web:' . substr($this->user_name, 0, 16);
        $thisDate = date('n/j/Y');
        $this->addField('created_by', $thisUser, TRUE);
        $this->addField('created_date', $thisDate, TRUE);
      }
      //Prepare the SET portion of the update statement.
      $varFields = $varValues = '';
      foreach ($this->update_fields as $field => $value) {
        $varFields .= $field . ',';
        $varValues .= $value . ',';
      }
      //Remove the final comma from $varFields and $varValues.
      $varFields = substr($varFields, 0, strlen($varFields)-1);
      $varValues = substr($varValues, 0, strlen($varValues)-1);
      //Prepare the final insert statement.
      $this->sql = "INSERT INTO {$this->base_table} ({$varFields}) VALUES ({$varValues})";
      //Execute API call.
      $this->execute_api_call();
    }
  }

  /**
   * Executes a SQL UPDATE statement.
   */
  protected function execute_update() {
    if (($this->base_table) && (count($this->update_fields) > 0) && (count($this->conditions) > 0)) {
      //If not already included, add the currently logged-in website user and date as modified_by and modified_date.
      if (!array_key_exists('modified_by',$this->update_fields) || !array_key_exists('modified_date',$this->update_fields)) {
        $thisUser = 'web:' . substr($this->user_name, 0, 16);
        $thisDate = date('n/j/Y');
        $this->addField('modified_by', $thisUser, TRUE);
        $this->addField('modified_date', $thisDate, TRUE);
      }
      //Prepare the SET portion of the update statement.
      $varSet = '';
      foreach ($this->update_fields as $field => $value) {
        $varSet .= "{$field}={$value},";
      }
      //Remove the final comma from $varSet.
      $varSet = substr($varSet, 0, strlen($varSet)-1);
      //Prepare WHERE portion of update statement.
      $varWhere = $this->compileWhere();
      //Prepare the final update statement.
      if (!empty($varSet) && !empty($varWhere)) {
        $this->sql = "UPDATE {$this->base_table} SET {$varSet} WHERE {$varWhere}";
        //Execute API call.
        $this->execute_api_call();
      }
    }
  }

  /**
   * Executes a SQL statement passed into the query object from
   *   an external source.
   */
  protected function execute_pass() {
    $this->execute_api_call();
  }

  /**
   * Performs the actual call to the DonorPerfect API.
   */
  protected function execute_api_call() {
    // Make sure we have an $auth_string.
    if (empty($this->auth_string)) {
      $this->result_code = 403;
      $this->result = [];
      return;
    }
    // Prepare API call.
    $http_method = 'GET';
    if ($this->is_proc) {
      $this->params_string = $this->filterString($this->params_string);
      $this->api_call = self::DONORPERFECT_API_URL . "?action={$this->dp_action}&params={$this->params_string}&{$this->auth_string}";
    } elseif (!empty($this->sql)) {
      $this->sql = $this->filterString($this->sql);
      if (strlen($this->sql) <= 255) {
        $this->api_call = self::DONORPERFECT_API_URL . "?action={$this->sql}&{$this->auth_string}";
      } else {
        // Use POST rather than GET request, as SQL statements can become
        // too long to fit in the url.
        $http_method = 'POST';
        $this->api_call = self::DONORPERFECT_API_URL . '?' . $this->auth_string;
        $this->setHttpOption('form_params', ['action' => $this->sql]);
      }
    } else {
      $this->api_call = '';
    }

    // Replace spaces in $this->api_call with %20.
    $this->api_call = str_replace(' ', '%20', $this->api_call);

    // Execute API call.
    if (!empty($this->api_call)) {
      $dpoReturn = \Drupal::httpClient()->request($http_method, $this->api_call, $this->http_request_options);
      $this->result_code = (int)$dpoReturn->getStatusCode();
    }

    // Clear out the result array.
    $this->result = [];

    // Parse results.
    if ($this->result_code === 200) {
      $result = $dpoReturn->getBody()->getContents();
      $thisResult = new \SimpleXMLElement($result);

      foreach ($thisResult->record as $rec) {
        $thisRow = [];
        foreach($rec->field as $thisField) {
          $fieldName = (string) $thisField['name'];
          $thisRow[$fieldName] = (string) $thisField['value'];
        }
        $this->result[] = $thisRow;
      }

      // Reconfigure the result array if we used one of the DonorPerfect "save" predefined procedures.
      if ($this->dp_action == 'dp_savedonor') {
        $temp = $this->result;
        $this->result = [];
        $this->result['donor_id'] = $temp[0][''];
      } elseif ($this->dp_action == 'dp_savegift') {
        $temp = $this->result;
        $this->result = [];
        $this->result['gift_id'] = $temp[0][''];
      } elseif ($this->dp_action == 'dp_savecontact') {
        $temp = $this->result;
        $this->result = [];
        $this->result['contact_id'] = $temp[0][''];
      } elseif ($this->dp_action == 'dp_saveotherinfo') {
        $temp = $this->result;
        $this->result = [];
        $this->result['other_id'] = $temp[0][''];
      }
    } else {
      // Problem getting data from DonorPerfect API.
      // Throw an exception?
    }
  }

  /**
   * Intermediary function for predefined procedures to ensure that all
   *   parameters are properly populated before being sent to the
   *   DonorPerfect API.
   */
  protected function execute_predefined_proc() {
    $thisParams = '';
    // Add the currently logged-in website user as the user_id so that it will be recorded in DP.
    $thisUser = 'web:' . substr($this->user_name, 0, 16);
    $this->addField('user_id', $thisUser, TRUE);

    foreach ($this->proc_fields as $thisField => $thisType) {
      if (array_key_exists($thisField, $this->params)) {
        $thisParams .= $this->params[$thisField] . ',';
      } else {
        // Set up default values for fields that don't allow null.
        if (($this->type === 'savedonor') && ($thisField == 'donor_id')) {
          $thisParams .= '0,';
        } elseif (($this->type === 'savedonor') && ($thisField === 'nomail')) {
          $thisParams .= "'N',";
        } elseif (($this->type === 'savegift') && ($thisField === 'gift_id')) {
          $thisParams .= '0,';
        } elseif (($this->type === 'savegift') && ($thisField === 'amount')) {
          $thisParams .= '0,';
        } elseif (($this->type === 'savegift') && ($thisField === 'batch_no')) {
          $thisParams .= '0,';
        } elseif (($this->type === 'savegift') && ($thisField === 'record_type')) {
          $thisParams .= "'G',";
        } elseif (($this->type === 'savegift') && ($thisField === 'split_gift')) {
          $thisParams .= "'N',";
        } elseif (($this->type === 'savegift') && ($thisField === 'pledge_payment')) {
          $thisParams .= "'N',";
        } elseif (($this->type === 'savegift') && ($thisField === 'nocalc')) {
          $thisParams .= "'N',";
        } elseif (($this->type === 'savegift') && ($thisField === 'receipt')) {
          $thisParams .= "'N',";
        } elseif (($this->type === 'saveotherinfo') && ($thisField === 'other_id')) {
          $thisParams .= '0,';
        } elseif (($this->type === 'savecontact') && ($thisField === 'contact_id')) {
          $thisParams .= '0,';
        } else { 
          $thisParams .= 'null,';
        }
      }
    }

    // Remove the final comma from $thisParams.
    $thisParams = substr($thisParams,0,strlen($thisParams)-1);

    $this->params_string = $thisParams; 
    $this->execute_api_call();
  }

  /**
   * Executes the DonorSearch predefined procedure.
   */
  protected function execute_donorsearch() {
    $this->dp_action = 'dp_donorsearch';
    $this->execute_predefined_proc();
  }

  /**
   * Executes the SaveDonor predefined procedure.
   */
  protected function execute_savedonor() {
    // First of all, determine whether we're creating a new donor or updating an existing one.
    // If creating a new donor, then we'll use DP's predefined procedure.
    // If updating an existing donor, then we'll create custom update sql statements (because the predefined proc stinks for updating).

    if (!array_key_exists('donor_id', $this->params) || ($this->params['donor_id'] == 0)) {
      // Ok, we're creating a new donor, so let's use the predefined procedure.
      $this->dp_action = 'dp_savedonor';
      $this->execute_predefined_proc();

      // Now we have created the new record and have the new donor_id in $this->result.
      // But maybe additional parameters were passed to us that were not saved to the new donor's record by the predefined proc.
      // Therefore, we're going to strip $this->params of all fields saved by the predefined proc and then add the new donor_id.
      // Then we will let the next part of this function update the new record with any additional fields that might be in $this->params.

      foreach ($this->proc_fields as $field => $type) {
        if (array_key_exists($field,$this->params)) {
          unset($this->params[$field]);
        }
      }
      $this->params['donor_id'] = $this->result['donor_id'];
    }

    // Now we are going to update an existing record. We need to have a donor_id and at least one more parameter.
    if (array_key_exists('donor_id', $this->params) && ($this->params['donor_id'] > 0) && (count($this->params) > 1)) {
      // Since we are working with a donor record, the two DonorPerfect tables we are concerned about are dp and dpudf.
      // Create a DPQuery update object for each of these two tables.
      $thisDp = new DPQuery('update','dp');
      $thisDpudf = new DPQuery('update','dpudf');

      // Add the currently logged-in website user and date as modified_by and modified_date.
      $thisUser = 'web:' . substr($this->user_name, 0, 16);
      $thisDate = date('n/j/Y');
      $thisDp->addField('modified_by', $thisUser, TRUE);
      $thisDp->addField('modified_date', $thisDate, TRUE);

      // Now pass all the params to both objects.
      // Let the object figure out whether or not each parameter is actually a field in its corresponding table.
      foreach ($this->params as $field => $value) {
        if ($field == 'donor_id') {
          //Donor_id is actually the condition rather than a field to be updated.
          $thisDp->addCondition('donor_id=' . $value);
          $thisDpudf->addCondition('donor_id=' . $value);
        } else {
          $thisDp->addField($field,$value);
          $thisDpudf->addField($field,$value);
        }
      }

      // Execute the updates on both objects.
      $thisDp->execute();
      $thisDpudf->execute();
    }
  }

  /**
   * Executes the Gifts predefined procedure.
   */
  protected function execute_gifts() {
    $this->dp_action = 'dp_gifts';
    $this->execute_predefined_proc();
  }

  /**
   * Executes the SaveGift predefined procedure.
   */
  protected function execute_savegift() {
    // First of all, determine whether we're creating a new gift or updating an existing one.
    // If creating a new gift, then we'll use DP's predefined procedure.
    // If updating an existing gift, then we'll create custom update sql statements (because the predefined proc stinks for updating).

    if (!array_key_exists('gift_id', $this->params) || ((int)$this->params['gift_id'] === 0)) {
      // Ok, we're creating a new gift, so let's use the predefined procedure.
      $this->dp_action = 'dp_savegift';
      $this->execute_predefined_proc();

      // Now we have created the new record and have the new gift_id in $this->result.
      // But maybe additional parameters were passed to us that were not saved to the new gift's record by the predefined proc.
      // Therefore, we're going to strip $this->params of all fields saved by the predefined proc and then add the new gift_id.
      // Then we will let the next part of this function update the new record with any additional fields that might be in $this->params.

      foreach ($this->proc_fields as $field => $type) {
        if (array_key_exists($field,$this->params)) {
          unset($this->params[$field]);
        }
      }
      $this->params['gift_id'] = $this->result['gift_id'];
    }

    // Now we are going to update an existing record. We need to have a gift_id and at least one more parameter.
    if (array_key_exists('gift_id', $this->params) && ($this->params['gift_id'] > 0) && (count($this->params) > 1)) {
      // Since we are working with a gift record, the two DonorPerfect tables we are concerned about are dpgift and dpgiftudf.
      // Create a DPQuery update object for each of these two tables.
      $thisDpgift = new DPQuery('update','dpgift');
      $thisDpgiftudf = new DPQuery('update','dpgiftudf');

      // Add the currently logged-in website user and date as modified_by and modified_date.
      $thisUser = 'web:' . substr($this->user_name, 0, 16);
      $thisDate = date('n/j/Y');
      $thisDpgift->addField('modified_by', $thisUser, TRUE);
      $thisDpgift->addField('modified_date', $thisDate, TRUE);

      // Now pass all the params to both objects.
      // Let the object figure out whether or not each parameter is actually a field in its corresponding table.
      foreach ($this->params as $field => $value) {
        if ($field == 'gift_id') {
          // Gift_id is actually the condition rather than a field to be updated.
          $thisDpgift->addCondition('gift_id=' . $value);
          $thisDpgiftudf->addCondition('gift_id=' . $value);
        } else {
          $thisDpgift->addField($field,$value);
          $thisDpgiftudf->addField($field,$value);
        }
      }
      // Execute the updates on both objects.
      $thisDpgift->execute();
      $thisDpgiftudf->execute();
    }
  }

  /**
   * Executes the SaveOtherInfo predefined procedure.
   */
  protected function execute_saveotherinfo() {
    // First of all, determine whether we're creating a new record or updating an existing one.
    // If creating a new record, then we'll use DP's predefined procedure.
    // If updating an existing record, then we'll create custom update sql statements (because the predefined proc stinks for updating).

    if (!array_key_exists('other_id', $this->params) || ($this->params['other_id'] == 0)) {
      // Ok, we're creating a new record, so let's use the predefined procedure.
      $this->dp_action = 'dp_saveotherinfo';
      $this->execute_predefined_proc();

      // Now we have created the new record and have the new other_id in $this->result.
      // But maybe additional parameters were passed to us that were not saved to the new record by the predefined proc.
      // Therefore, we're going to strip $this->params of all fields saved by the predefined proc and then add the new other_id.
      // Then we will let the next part of this function update the new record with any additional fields that might be in $this->params.

      foreach ($this->proc_fields as $field => $type) {
        if (array_key_exists($field,$this->params)) {
          unset($this->params[$field]);
        }
      }
      $this->params['other_id'] = $this->result['other_id'];
    }


    // Now we are going to update an existing record. We need to have a other_id and at least one more parameter.
    if (array_key_exists('other_id', $this->params) && ($this->params['other_id'] > 0) && (count($this->params) > 1)) {
      // Since we are working with an otherinfo record, the two DonorPerfect tables we are concerned about are dpotherinfo and dpotherinfoudf.
      // Create a DPQuery update object for each of these two tables.
      $thisDp = new DPQuery('update','dpotherinfo');
      $thisDpudf = new DPQuery('update','dpotherinfoudf');

      // Add the currently logged-in website user and date as modified_by and modified_date.
      $thisUser = 'web:' . substr($this->user_name, 0, 16);
      $thisDate = date('n/j/Y');
      $thisDp->addField('modified_by', $thisUser, TRUE);
      $thisDp->addField('modified_date', $thisDate, TRUE);

      // Now pass all the params to both objects.
      // Let the object figure out whether or not each parameter is actually a field in its corresponding table.
      foreach ($this->params as $field => $value) {
        if ($field == 'other_id') {
          // Other_id is actually the condition rather than a field to be updated.
          $thisDp->addCondition('other_id=' . $value);
          $thisDpudf->addCondition('other_id=' . $value);
        } else {
          $thisDp->addField($field,$value);
          $thisDpudf->addField($field,$value);
        }
      }

      // Execute the updates on both objects.
      $thisDp->execute();
      $thisDpudf->execute();
    }
  }

  /**
   * Executes the SaveContact predefined procedure.
   */
  protected function execute_savecontact() {
    // First of all, determine whether we're creating a new record or updating an existing one.
    // If creating a new record, then we'll use DP's predefined procedure.
    // If updating an existing record, then we'll create custom update sql statements (because the predefined proc stinks for updating).

    if (!array_key_exists('contact_id', $this->params) || ($this->params['contact_id'] == 0)) {
      // Ok, we're creating a new record, so let's use the predefined procedure.
      $this->dp_action = 'dp_savecontact';
      $this->execute_predefined_proc();

      // Now we have created the new record and have the new contact_id in $this->result.
      // But maybe additional parameters were passed to us that were not saved to the new record by the predefined proc.
      // Therefore, we're going to strip $this->params of all fields saved by the predefined proc and then add the new contact_id.
      // Then we will let the next part of this function update the new record with any additional fields that might be in $this->params.

      foreach ($this->proc_fields as $field => $type) {
        if (array_key_exists($field,$this->params)) {
          unset($this->params[$field]);
        }
      }
      $this->params['contact_id'] = $this->result['contact_id'];
    }

    // Now we are going to update an existing record. We need to have a contact_id and at least one more parameter.
    if (array_key_exists('contact_id', $this->params) && ($this->params['contact_id'] > 0) && (count($this->params) > 1)) {
      // Since we are working with a conact record, the two DonorPerfect tables we are concerned about are dpcontact and dpcontactudf.
      // Create a DPQuery update object for each of these two tables.
      $thisDp = new DPQuery('update','dpcontact');
      $thisDpudf = new DPQuery('update','dpcontactudf');

      // Add the currently logged-in website user and date as modified_by and modified_date.
      $thisUser = 'web:' . substr($this->user_name, 0, 16);
      $thisDate = date('n/j/Y');
      $thisDp->addField('modified_by', $thisUser, TRUE);
      $thisDp->addField('modified_date',$thisDate, TRUE);

      // Now pass all the params to both objects.
      // Let the object figure out whether or not each parameter is actually a field in its corresponding table.
      foreach ($this->params as $field => $value) {
        if ($field == 'contact_id') {
          // Contact_id is actually the condition rather than a field to be updated.
          $thisDp->addCondition('contact_id=' . $value);
          $thisDpudf->addCondition('contact_id=' . $value);
        } else {
          $thisDp->addField($field,$value);
          $thisDpudf->addField($field,$value);
        }
      }

      // Execute the updates on both objects.
      $thisDp->execute();
      $thisDpudf->execute();
    }
  }

  /**
   * Executes the DelFlags predefined procedure.
   */
  protected function execute_delflags() {
    $this->dp_action = 'dp_delflags_xml';
    $this->execute_predefined_proc();
  }

  /**
   * Executes the SaveFlag predefined procedure.
   */
  protected function execute_saveflag() {
    $this->dp_action = 'dp_saveflag_xml';
    $this->execute_predefined_proc();
  }

  /**
   * Executes the SaveUdf predefined procedure.
   */
  protected function execute_saveudf() {
    $this->dp_action = 'dp_save_udf_xml';
    $this->execute_predefined_proc();
  }

  /**
   * Get the fields for a specific predefined procedure as defined by
   *   DonorPerfect.
   * 
   * @return array
   *   Key is the field name, and value is the data type.
   *   (n = numeric, d=date, and c=character)
   */
  protected function getProcFields() {
    $proc_fields = [];
    switch ($this->type) {
      case 'delflags':
        $proc_fields = [
          'donor_id' => 'n',
          'user_id' => 'c',
        ];
        break;
      case 'donorsearch':
        $proc_fields = [
          'donor_id' => 'n',
          'last_name' => 'c',
          'first_name' => 'c',
          'opt_line' => 'c',
          'address' => 'c',
          'city' => 'c',
          'state' => 'c',
          'zip' => 'c',
          'country' => 'c',
          'filter_id' => 'n',
          'user_id' => 'c',
        ];
        break;
      case 'gifts':
        $proc_fields = [
          'donor_id' => 'n',
        ];
        break;
      case 'savecontact':
        $proc_fields = [
          'contact_id' => 'n', // Enter 0 to create a new record, or an existing contact_id to update a record.
          'donor_id' => 'n',
          'activity_code' => 'c',
          'mailing_code' => 'c',
          'by_whom' => 'c',
          'contact_date' => 'd',
          'due_date' => 'd',
          'due_time' => 'c',
          'completed_date' => 'd',
          'comment' => 'c',
          'document_path' => 'c',
          'user_id' => 'c',
        ];
        break;
      case 'savedonor':
        $proc_fields = [
          'donor_id' => 'n',  // Must be 0 to create a new record, or an existing donor_id to update a record.
          'first_name' => 'c',
          'last_name' => 'c',
          'middle_name' => 'c',
          'suffix' => 'c',
          'title' => 'c',
          'salutation' => 'c',
          'prof_title' => 'c',
          'opt_line' => 'c',
          'address' => 'c',
          'address2' => 'c',
          'city' => 'c',
          'state' => 'c',
          'zip' => 'c',
          'country' => 'c',
          'address_type' => 'c',
          'home_phone' => 'c',
          'business_phone' => 'c',
          'fax_phone' => 'c',
          'mobile_phone' => 'c',
          'email' => 'c',
          'org_rec' => 'c',
          'donor_type' => 'c',
          'nomail' => 'c',
          'nomail_reason' => 'c',
          'narrative' => 'c',
          'user_id' => 'c',
        ];
        break;
      case 'saveflag':
        $proc_fields = [
          'donor_id' => 'n',
          'flag' => 'c',
          'user_id' => 'c',
        ];
        break;
      case 'savegift':
        $proc_fields = [
          'gift_id' => 'n', // Enter 0 to create new gift record, or an existing gift_id to update a record.
          'donor_id' => 'n',
          'record_type' => 'c',
          'gift_date' => 'd',
          'amount' => 'n',
          'gl_code' => 'c',
          'solicit_code' => 'c',
          'sub_solicit_code' => 'c',
          'gift_type' => 'c',
          'split_gift' => 'c',
          'pledge_payment' => 'c',
          'reference' => 'c',
          'memory_honor' => 'c',
          'gfname' => 'c',
          'glname' => 'c',
          'fmv' => 'n',
          'batch_no' => 'n',
          'gift_narrative' => 'c',
          'ty_letter_no' => 'c',
          'glink' => 'n',
          'plink' => 'n',
          'nocalc' => 'c',
          'receipt' => 'c',
          'old_amount' => 'n',
          'user_id' => 'c',
          'campaign' => 'c',
          'membership_type' => 'c',
          'membership_level' => 'c',
          'membership_enr_date' => 'd',
          'membership_exp_date' => 'd',
          'membership_link_id' => 'n',
          'address_id' => 'n',
          'rcpt_status' => 'c',
          'rcpt_type' => 'c',
          'membership_code' => 'c',
          'contact_id' => 'n',
          'receipt_delivery_g' => 'c',
        ];
        break;
      case 'saveotherinfo':
        $proc_fields = [
          'other_id' => 'n', // Enter 0 to create a new record, or an existing other_id to update a record.
          'donor_id' => 'n',
          'other_date' => 'd',
          'comments' => 'c',
          'user_id' => 'c',
        ];
        break;
      case 'saveudf':
        $proc_fields = [
          'matching_id' => 'n',
          'field_name' => 'c',
          'data_type' => 'c',
          'char_value' => 'c',
          'date_value' => 'd',
          'number_value' => 'n',
          'user_id' => 'c',
        ];
        break;
    }
    return $proc_fields;
  }

  /**
   * Given a table name, return the field that should be base_field for a query.
   *
   * @param string $table
   *
   * @return string
   */
  protected function getBaseField(string $table) {
    $output = '';
    switch ($table) {
      case 'dp':
      case 'dpudf':
      case 'dpflags':
        $output = 'donor_id';
        break;
      case 'dpgift':
      case 'dpgiftudf':
        $output = 'gift_id';
        break;
      case 'dpaddress':
      case 'dpaddressmailings':
        $output = 'address_id';
        break;
      case 'dplink':
        $output = 'link_id';
        break;
      case 'dpotherinfo':
      case 'dpotherinfoudf':
        $output = 'other_id';
        break;
      case 'dpcontact':
      case 'dpcontactudf':
        $output = 'contact_id';
        break;
      case 'dpusermultivalues':
      case 'dpbatch':
        $output = 'id';
        break;
      case 'dpcodes':
        $output = 'code';
        break;
    }
    return $output;
  }

  /**
   * Executes a series of queries to retrieve all data from a specified
   *   table in DonorPerfect.
   */
  protected function execute_retrieveAllTableData() {
    if ($this->base_table) {
      $thisResult = [];
      $query = new DPQuery();

      // Get the count of the number of records in the table.
      $tempSql = "SELECT count(*) AS count FROM {$this->base_table}";
      $query->reset('pass',$tempSql);
      $query->execute();
      $thisCount = $query->result[0]['count'];
      if (!($thisCount > 0)) {
        $this->result_code = 404;
        $this->result = [];
        return;
      }

      // Get a column name to order by.
      $tempSql = "SELECT TOP 1 * FROM {$this->base_table}";
      $query->reset('pass',$tempSql);
      $query->execute();
      if (count($query->result) == 1) {
        foreach ($query->result AS $field => $value) {
          $tempKeys = array_keys($value);
          $thisOrderBy = $tempKeys[0];
          break;
        }
      } else {
        $this->result_code = 404;
        $this->result = [];
        return;
      }

      // Retrieve the records in groups of 400 because DonorPerfect
      // limits the number of records that we can retrieve at one time. 
      $groupCount = ($thisCount % 400 == 0) ? $thisCount/400 : intval($thisCount/400) + 1;

      for ($x=0; $x <= $groupCount-1; $x++) {
        $y = ($x * 400) + 1;
        $z = 400 + ($x * 400);

        $tempSql = "WITH mytable AS (SELECT *,ROW_NUMBER() OVER (order by {$thisOrderBy}) AS RowNumber FROM {$this->base_table}) SELECT * FROM myTable WHERE (RowNumber BETWEEN {$y} AND {$z})";
        $query->reset('pass',$tempSql);
        $query->execute();

        $thisResult = array_merge($thisResult, $query->result);
      }

      $this->result_code = 200;
      $this->result = $thisResult;
      return;
    } else {
      $this->result_code = 404;
      $this->result = [];
      return;
    }
  }

 /**
  * Returns a single value from the result array.
  *
  * @return mixed
  */
  public function fetchField() {
    if (is_array($this->result) && isset($this->result[0])) {
      if (is_array($this->result[0])) {
        return current($this->result[0]);
      } else {
        return $this->result[0];
      }
    } else {
      return NULL;
    }
  }

  /**
   * Return $this->result as an array of objects as required by Drupal Views.
   *
   * @return mixed
   */
  public function resultObjects() {
    $return = [];
    foreach ($this->result as $key => $value) {
      if (is_array($value)) {
        $temp = new stdClass();
        foreach ($value as $tempKey => $tempValue) {
          $temp->{$tempKey} = $tempValue;
        }
      } else {
        $temp = $value;
      }
      $return[$key] = $temp;
    }
    return $return;
  }

  /**
   * Helper function to add single quotes around a value for use in a sql query.
   *
   * @param string $value
   *   The value around which to add single quotes.
   *
   * @return string
   */
  public function addSingleQuotes(string $value) {
    //Escape any single quotes already in $value.
    $value = str_replace("'", "''", $value);
    return "'{$value}'";
  }

  /**
   * Helper function to remove characters from a string that cause problems for DPO API.
   *
   * @param string $value
   *   The value in which to replace troublesome characters.
   *
   * @return string
   */
  public function filterString(string $value) {
    $value = str_replace('#', 'Number ', $value);
    $value = str_replace('&', 'and', $value);
    return $value;
  }

 /**
  * This function resets all of the DPQuery object's properties back to their default values.
  * Returns the object to enable daisy chaining.
  *
  * @param string $type
  *   The type of query to reset the object to.
  * @param mixed $optional
  *   Optional parameter to pass into the object constructor.
  *
  * @return \Drupal\donorperfect\DPQuery $this
  */
  public function reset(string $type = '', mixed $optional = null) {
    $this->__construct($type, $optional);
    $this->tables_fields = [];
    $this->conditions = [];
    $this->order_by = [];
    $this->order_by_string = '';
    $this->update_fields = [];
    $this->api_call = '';
    $this->dp_action = '';
    $this->params = [];
    $this->params_string = '';
    $this->limit = 0;
    $this->offset = 0;
    $this->result_code = '';
    $this->result = [];
    $this->http_request_options = ['timeout' => 30];
    return $this;
  }

  /**
   * This module stores a local cache containing information about different aspects 
   *   of DonorPerfect (field definitions, code definitions, etc). This method
   *   refreshes that cache.
   */
  public function refreshCache() {
    $this->loadDpcodes();
    $this->loadDpmultivalues();
    $this->loadDptables();
  }

  /**
   * Part of the process to refreshCache(). Loads table and field definitions
   *   from DonorPerfect to store in the local cache.
   */
  protected function loadDptables() {
    $theseFields = \Drupal::config('donorperfect.cache')->get('dptables') ?? [];
    $theseCodes = \Drupal::config('donorperfect.cache')->get('dpcodes') ?? [];
    // Get a list of calculated fields so that we can determine which fields are not editable.
    $theseCalculated = [];
    $query2 = new DPQuery('pass', 'SELECT field_to_update FROM dp_calculated_fields');
    $query2->execute();
    foreach ($query2->result as $thisRow) {
      $temp = strtolower($thisRow['field_to_update']);
      $theseCalculated[] = $temp;
    }
    // Get all field data from sd_field.
    $query2->reset('retrieveAllTableData','sd_field');
    $query2->execute();
    $tempFields = $query2->result;
    // Get all table_names from sd_section.
    $query2->reset('pass', 'SELECT section_id,table_name FROM sd_section');
    $query2->execute();
    $tempSections = [];
    foreach ($query2->result as $section) {
      $tempSections['s' . $section['section_id']] = $section['table_name'];
    }
    // Re-key field info by table_name and field_name.
    $theseFieldInfo = [];
    foreach ($tempFields as $tempField) {
      if (!empty($tempField['field_name']) && !empty($tempField['section_id']) && ($tempTable = $tempSections['s' . $tempField['section_id']])) {
        $tempTable = strtolower($tempTable);
        if ($tempTable === 'bio') $tempTable = 'donor';
        $theseFieldInfo[$tempTable][strtolower($tempField['field_name'])] = $tempField;
      }
    }
    // Get info for all fields in tables included in $this->avail_tables.
    foreach ($this->avail_tables as $thisTable) {
      //Determine the table_name from sd_section that corresponds to this table.
      $sdTable = '';
      switch ($thisTable) {
        case 'dp':
        case 'dpudf':
          $sdTable = 'donor';
          break;
        case 'dpgift':
        case 'dpgiftudf':
          $sdTable = 'gift';
          break;
        case 'dpaddress':
          $sdTable = 'address';
          break;
        case 'dpaddressmailings':
          $sdTable = 'dpaddressmailings';
          break;
        case 'dplink':
          $sdTable = 'dplink';
          break;
        case 'dpotherinfo':
        case 'dpotherinfoudf':
          $sdTable = 'other';
          break;
        case 'dpcontact':
        case 'dpcontactudf':
          $sdTable = 'contact';
          break;
        case 'dpcodes':
          $sdTable = 'dpcodes';
          break;
        case 'dpflags':
          $sdTable = 'dpflags';
          break;
      }
      if (empty($sdTable)) continue;
      // Get field names for $thisTable.
      $query = new DPQuery('pass', "SELECT TOP 1 * FROM {$thisTable}");
      $query->execute();
      $tempFields = [];
      if ((count($query->result) === 1) && is_array($query->result[0])) {
        foreach ($query->result[0] as $key => $value) {
          $tempFields[] = strtolower($key);
        }
      }
      // Get info for each of the fields in $thisTable and add to final $theseFields array.
      foreach ($tempFields as $thisField) {
        if ($thisFieldInfo = NestedArray::getValue($theseFieldInfo, [$sdTable, $thisField])) {
          $thisDataType = strtoupper($thisFieldInfo['data_type']);
          $thisSqlDataType = '';
          $thisFormType = $thisNumericScale = $thisMulti = FALSE;
          switch ($thisDataType) {
            case 'C':
              $thisQuote = TRUE;
              $thisSqlDataType = 'varchar';
              switch ($thisFieldInfo['display_type']) {
                case 'L': //DonorPerfect code for textarea.
                  $thisEntityDataType = 'string_long';
                  $thisFormType = 'textarea';
                  break;
                case 'C': //DonorPerect code for boolean checkbox.
                  $thisEntityDataType = 'boolean';
                  $thisFormType = 'checkbox';
                  break;
                default:
                  $thisEntityDataType = 'string';
                  $thisFormType = 'textfield';
                  break;
              }
              break;
            case 'D':
              $thisQuote = TRUE;
              $thisSqlDataType = 'datetime';
              $thisEntityDataType = 'datetime';
              break;
            case 'M':
              $thisQuote = FALSE;
              $thisSqlDataType = 'money';
              $thisEntityDataType = 'decimal';
              $thisNumericScale = 2;
              break;
            case 'N':
              $thisQuote = FALSE;
              $thisSqlDataType = 'int';
              $thisEntityDataType = 'integer';
              break;
            default:
              $thisQuote = TRUE;
              $thisSqlDataType = 'varchar';
              $thisEntityDataType = 'string';
              break;
          }
          // Check if this is a coded field.
          $thisCode = FALSE;
          $theseOptions = [];
          if (array_key_exists($thisField, $theseCodes)) {
            $thisCode = TRUE;
            $theseOptions = $theseCodes[$thisField];
            $thisEntityDataType = 'list_string';
          }
          // Check if this is a calculated field.
          $thisCalculated = FALSE;
          if (in_array($thisField, $theseCalculated)) {
            $thisCalculated = TRUE;
          }
          $theseFields[$thisTable][$thisField] = [
            'dp_table' => $thisTable,
            'dp_field' => $thisField,
            'prompt' => $thisFieldInfo['field_prompt'],
            'mysql_data_type' => $thisSqlDataType,
            'max_length' => $thisFieldInfo['field_length'],
            'entity_data_type' => $thisEntityDataType,
            'form_type' => $thisFormType,
            'quote' => $thisQuote,
            'code' => $thisCode,
            'options' => $theseOptions,
            'multi' => $thisMulti,
            'calculated' => $thisCalculated,
            'editable' => !$thisCalculated,
            'decimals' => $thisNumericScale,
          ];
        }
      }
    }
    // Add in the multivalues so that they appear as regular fields.
    $theseMulti = \Drupal::config('donorperfect.cache')->get('dpmultivalues') ?? [];
    foreach ($theseMulti as $thisMulti) {
      $thisTable = $sdTable = '';
      switch ($thisMulti['screen']) {
        case 'main':
        case 'bio':
          $thisTable = 'dpudf';
          $sdTable = 'donor';
          break;
        case 'gift':
        case 'pledge':
          $thisTable = 'dpgiftudf';
          $sdTable = 'gift';
          break;
        case 'address':
          $thisTable = $sdTable = 'dpaddressmailings';
          break;
        case 'contact':
          $thisTable = 'dpcontactudf';
          $sdTable = 'contact';
          break;
        case 'other':
          $thisTable = 'dpotherudf';
          $sdTable = 'other';
          break;
      }
      if (!empty($sdTable) && ($thisFieldInfo = NestedArray::getValue($theseFieldInfo, [$sdTable, $thisMulti['field_name']]))) {
        $theseFields[$thisTable][$thisMulti['field_name']] = [
          'dp_table' => $thisTable,
          'dp_field' => $thisMulti['field_name'],
          'prompt' => $thisFieldInfo['field_prompt'],
          'mysql_data_type' => 'varchar',
          'max_length' => $thisFieldInfo['field_length'],
          'entity_data_type' => 'text',
          'form_type' => 'textfield',
          'quote' => TRUE,
          'code' => TRUE,
          'options' => !empty($theseCodes[$thisMulti['field_name']]) ? $theseCodes[$thisMulti['field_name']] : [],
          'multi' => TRUE,
          'calculated' => FALSE,
          'editable' => TRUE,
          'decimals' => FALSE,
        ];
      }
    }
    // Add the special-case flags field to the dp table.
    $theseFields['dp']['flag'] = [
      'dp_table' => 'dp',
      'dp_field' => 'flag',
      'prompt' => 'Flag',
      'mysql_data_type' => 'varchar',
      'max_length' => 30,
      'entity_data_type' => 'list_string',
      'form_type' => 'textfield',
      'quote' => TRUE,
      'code' => TRUE,
      'options' => !empty($theseCodes['flag']) ? $theseCodes['flag'] : [],
      'multi' => TRUE,
      'calculated' => FALSE,
      'editable' => TRUE,
      'decimals' => FALSE,
    ];
    $theseFields['dpflags']['donor_id'] = [
      'dp_table' => 'dpflags',
      'dp_field' => 'donor_id',
      'prompt' => 'Donor ID',
      'mysql_data_type' => 'int',
      'max_length' => 100,
      'entity_data_type' => 'integer',
      'form_type' => FALSE,
      'quote' => FALSE,
      'code' => FALSE,
      'options' => [],
      'multi' => FALSE,
      'calculated' => FALSE,
      'editable' => TRUE,
      'decimals' => FALSE,
    ];
    // Sort the table and field arrays alphabetically.
    ksort($theseFields);
    foreach ($theseFields as &$thisTable) {
      ksort($thisTable);
    }
    \Drupal::configFactory()
      ->getEditable('donorperfect.cache')
      ->set('dptables', $theseFields)
      ->save();
  }

  /**
   * Part of the refreshCache() process. Loads all codes as configured
   *   in the 'Code Maintenance' page in DonorPerfect.
   */
  protected function loadDpcodes() {
    $theseCodes = [];
    $query = new DPQuery();

    $tempSql = "SELECT count(code) AS count FROM dpcodes WHERE NOT field_name='w4w_walker'";
    $query->reset('pass',$tempSql);
    $query->execute();
    $codeCount = $query->result[0]['count'];

    // Retrieve the codes in groups of 400 because DonorPerfect limits the number of records that we can retrieve at one time. 
    $groupCount = ($codeCount % 400 == 0) ? $codeCount/400 : intval($codeCount/400) + 1;

    for ($x=0; $x <= $groupCount-1; $x++) {
      $y = ($x * 400) + 1;
      $z = 400 + ($x * 400);

      $tempSql = "WITH mytable AS (SELECT *,ROW_NUMBER() OVER (order by field_name,description) AS RowNumber FROM dpcodes WHERE NOT field_name='w4w_walker') SELECT code,description,field_name,inactive FROM myTable WHERE (RowNumber BETWEEN {$y} AND {$z})";
      $query->reset('pass',$tempSql);
      $query->execute();

      foreach ($query->result AS $thisRow) {
//        $theseCodes[strtolower($thisRow['field_name'])][strtolower($thisRow['code'])] = $thisRow['description'];
        $theseCodes[strtolower($thisRow['field_name'])][$thisRow['code']] = [
          'dp_field' => strtolower($thisRow['field_name']),
          'code' => $thisRow['code'],
          'description' => $thisRow['description'],
          'active' => ($thisRow['inactive'] == 'N') ? TRUE : FALSE,
        ];
      }
    }

    \Drupal::configFactory()
      ->getEditable('donorperfect.cache')
      ->set('dpcodes', $theseCodes)
      ->save();
  }

  /**
   * Part of the refreshCache() process. Loads information about multi-value
   *   fields in DonorPerfect to store in the local cache.
   */
  protected function loadDpmultivalues() {
    $query = new DPQuery();
    $theseMulti = [];

    $tempSql = "SELECT table_name,field_name FROM dpuserfields WHERE display_type='M'";
    $query->reset('pass',$tempSql);
    $query->execute();

    foreach ($query->result as $thisRow) {
      $theseMulti[strtolower($thisRow['field_name'])] = [
        'screen' => strtolower($thisRow['table_name']),
        'field_name' => strtolower($thisRow['field_name']),
      ];
    }
 
    \Drupal::configFactory()
      ->getEditable('donorperfect.cache')
      ->set('dpmultivalues', $theseMulti)
      ->save();
  }

}
