<?php

namespace Drupal\donorperfect\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\donorperfect\FormElementBase;

/**
 * Provides a form element for DonorPerfect donor phone number.
 *
 * @FormElement("donorperfect_phone")
 */
class Phone extends FormElementBase {

  /**
   * Max length of specific fields as set by DonorPerfect.
   */
  const PHONE_MAXLENGTH = 40;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#default_value' => NULL,
      '#process' => [
        [$class, 'processElement'],
      ],
      '#pre_render' => [],
      '#after_build' => [],
      '#element_validate' => [
        [$class, 'validateElement'],
      ],
      '#theme_wrappers' => [],
      '#phone_type_default' => 'mobile', // mobile, home, business, or fax
      '#phone_type_access' => TRUE,      // whether or not user can access the phone_type dropdown
      '#phone_type_options' => [],       // values that should be included in the phone_type dropdown
      '#phone_mask' => '(999) 999-9999', // input mask for phone numbers
      '#phone_maxlength' => static::PHONE_MAXLENGTH,
    ];
  }

  /**
   * Processes the form element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function processElement(array $element, FormStateInterface $form_state, array &$complete_form) {
    $default_value = !empty($element['#default_value']) && is_array($element['#default_value']) ? $element['#default_value'] : [];
    if (empty($default_value['phone_type'])) $default_value['phone_type'] = $element['#phone_type_default'] ?? 'mobile';
    $element_token = static::getElementToken($element, $form_state);
    $element['element_token'] = [
      '#type' => 'hidden',
      '#default_value' => $element_token,
      '#attributes' => ['class' => ['donorperfect-element-token']],
    ];
    $element['#prefix'] = Markup::create('<div class="donorperfect-element-wrapper donorperfect-phone-wrapper" id="donorperfect-element-wrapper-' . $element_token . '" element-token="' . $element_token . '">');
    $element['#suffix'] = Markup::create('</div>');
    // Phone Type dropdown
    $phone_type_access = $element['#phone_type_access'] ?? TRUE;
    $phone_type_options = [];
    $phone_type_options_all = [
      'mobile' => t('Mobile Phone'),
      'home' => t('Home Phone'),
      'business' => t('Business Phone'),
      'fax' => t('Fax'),
    ];
    if (!empty($element['#phone_type_options']) && is_array($element['#phone_type_options'])) {
      foreach ($element['#phone_type_options'] as $type) {
        if (array_key_exists($type, $phone_type_options_all)) {
          $phone_type_options[$type] = $phone_type_options_all[$type];
        }
      }
    } else {
      $phone_type_options = $phone_type_options_all;
    }
    if (!array_key_exists($default_value['phone_type'], $phone_type_options)) {
      if (array_key_exists($default_value['phone_type'], $phone_type_options_all)) {
        $phone_type_options[$default_value['phone_type']] = $phone_type_options_all[$default_value['phone_type']];
      } else {
        $default_value['phone_type'] = '';
      }
    }
    if ($phone_type_access) {
      $element['phone_type'] = [
        '#type' => 'select',
        '#title' => t('Phone Type'),
        '#title_display' => 'invisible',
        '#options' => $phone_type_options,
        '#default_value' => $default_value['phone_type'],
        '#required' => $element['#required'] ?? FALSE,
        '#attributes' => ['class' => ['phone-type-' . $element_token, 'dpfield'], 'token' => 'phone-type-' . $element_token, 'dpfield' => 'phone_type'],
        '#wrapper_attributes' => ['class' => ['phone-type-wrapper']],
      ];
    } else {
      $element['phone_type'] = [
        '#type' => 'hidden',
        '#default_value' => $default_value['phone_type'],
        '#attributes' => ['class' => ['phone-type-' . $element_token, 'dpfield'], 'token' => 'phone-type-' . $element_token, 'dpfield' => 'phone_type'],
      ];
    }
    // Phone textfields
    $element['mobile_phone'] = [
      '#type' => 'textfield',
      '#title' => t('Mobile Phone Number'),
      '#title_display' => 'invisible',
      '#default_value' => $default_value['mobile_phone'] ?? '',
      '#attributes' => ['class' => ['mobile-phone-' . $element_token, 'dpfield'], 'dpfield' => 'mobile_phone', 'placeholder' => t('Mobile phone number')],
      '#wrapper_attributes' => ['class' => ['moble-phone-wrapper']],
      '#maxlength' => $element['#phone_maxlength'] ?? static::PHONE_MAXLENGTH,
      '#states' => [
        'visible' => [
          ':input[token="phone-type-' . $element_token. '"]' => ['value' => 'mobile'],
        ]
      ],
    ];
    $element['home_phone'] = [
      '#type' => 'textfield',
      '#title' => t('Home Phone Number'),
      '#title_display' => 'invisible',
      '#default_value' => $default_value['home_phone'] ?? '',
      '#attributes' => ['class' => ['home-phone-' . $element_token, 'dpfield'], 'dpfield' => 'home_phone', 'placeholder' => t('Home phone number')],
      '#wrapper_attributes' => ['class' => ['home-phone-wrapper']],
      '#maxlength' => $element['#phone_maxlength'] ?? static::PHONE_MAXLENGTH,
      '#states' => [
        'visible' => [
          ':input[token="phone-type-' . $element_token. '"]' => ['value' => 'home'],
        ]
      ],
    ];
    $element['business_phone'] = [
      '#type' => 'textfield',
      '#title' => t('Business Phone Number'),
      '#title_display' => 'invisible',
      '#default_value' => $default_value['business_phone'] ?? '',
      '#attributes' => ['class' => ['business-phone-' . $element_token, 'dpfield'], 'dpfield' => 'business_phone', 'placeholder' => t('Business phone number')],
      '#wrapper_attributes' => ['class' => ['business-phone-wrapper']],
      '#maxlength' => $element['#phone_maxlength'] ?? static::PHONE_MAXLENGTH,
      '#states' => [
        'visible' => [
          ':input[token="phone-type-' . $element_token. '"]' => ['value' => 'business'],
        ]
      ],
    ];
    $element['fax_phone'] = [
      '#type' => 'textfield',
      '#title' => t('Fax Phone Number'),
      '#title_display' => 'invisible',
      '#default_value' => $default_value['fax_phone'] ?? '',
      '#attributes' => ['class' => ['fax-phone-' . $element_token, 'dpfield'], 'dpfield' => 'fax_phone', 'placeholder' => t('Fax phone number')],
      '#wrapper_attributes' => ['class' => ['fax-phone-wrapper']],
      '#maxlength' => $element['#phone_maxlength'] ?? static::PHONE_MAXLENGTH,
      '#states' => [
        'visible' => [
          ':input[token="phone-type-' . $element_token. '"]' => ['value' => 'fax'],
        ]
      ],
    ];
    if (!empty($element['#required'])) {
      $element['mobile_phone']['#states']['required'] = [':input[token="phone-type-' . $element_token. '"]' => ['value' => 'mobile']];
      $element['home_phone']['#states']['required'] = [':input[token="phone-type-' . $element_token. '"]' => ['value' => 'home']];
      $element['business_phone']['#states']['required'] = [':input[token="phone-type-' . $element_token. '"]' => ['value' => 'business']];
      $element['fax_phone']['#states']['required'] = [':input[token="phone-type-' . $element_token. '"]' => ['value' => 'fax']];
    }
    if (!empty($element['#phone_mask']) && is_string($element['#phone_mask'])) {
      $element['mobile_phone']['#attributes']['data-mask-value'] = $element['#phone_mask'];
      $element['home_phone']['#attributes']['data-mask-value'] = $element['#phone_mask'];
      $element['business_phone']['#attributes']['data-mask-value'] = $element['#phone_mask'];
      $element['fax_phone']['#attributes']['data-mask-value'] = $element['#phone_mask'];
    }
    // Add libraries.
    $element['#attached']['library'][] = 'donorperfect/form_element';
    return $element;
  }

  /**
   * Validates the form element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   */
  public static function validateElement(array $element, FormStateInterface $form_state, array $complete_form) {
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    $output = $element['#default_value'] = (array)$element['#default_value'];
    if (is_array($input)) {
      $output = $input;
    }
    return $output;
  }

}
