<?php

namespace Drupal\donorperfect\Entity\Query;

use Drupal\Core\Entity\Query\QueryException;
use Drupal\Core\Entity\Query\Sql\Query as SqlQuery;
use Drupal\donorperfect\DPQuery;

/**
 * The DonorPerfect entity query class.
 */
class Query extends SqlQuery {

  /**
   * Prepares the basic query with proper metadata/tags and base fields.
   *
   * @return $this
   *   Returns the called object.
   *
   * @throws \Drupal\Core\Entity\Query\QueryException
   *   Thrown if the base table does not exist.
   */
  protected function prepare() {
    if (!$base_table = $this->entityType->getBaseTable()) {
      throw new QueryException("No base table for " . $this->entityTypeId . ", invalid query.");
    }
    $simple_query = TRUE;
    $this->sqlQuery = $this->connection->select($base_table, 'base_table', ['conjunction' => $this->conjunction]);
    // Reset the tables structure, as it might have been built for a previous
    // execution of this query.
    $this->tables = NULL;
    $this->sqlQuery->addMetaData('entity_type', $this->entityTypeId);
    $id_field = $this->entityType->getKey('id');
    // Add the key field for fetchAllKeyed().
    $this->sqlFields["base_table.$id_field"] = ['base_table', $id_field];
    // Now add the value column for fetchAllKeyed(). This is always the
    // entity id.
    $this->sqlFields["base_table.$id_field" . '_1'] = ['base_table', $id_field];

    $this->accessCheck = FALSE;

    $this->sqlQuery->addTag('entity_query');
    $this->sqlQuery->addTag('entity_query_' . $this->entityTypeId);

    // Add further tags added.
    if (isset($this->alterTags)) {
      foreach ($this->alterTags as $tag => $value) {
        $this->sqlQuery->addTag($tag);
      }
    }

    // Add further metadata added.
    if (isset($this->alterMetaData)) {
      foreach ($this->alterMetaData as $key => $value) {
        $this->sqlQuery->addMetaData($key, $value);
      }
    }
    // This now contains first the table containing entity properties and
    // last the entity base table. They might be the same.
    $this->sqlQuery->addMetaData('all_revisions', FALSE);
    $this->sqlQuery->addMetaData('simple_query', $simple_query);

    return $this;
  }

  /**
   * Executes the query and returns the result.
   *
   * @return int|array
   *   Returns the query result as entity IDs.
   */
  protected function result() {
    $sql_query = $this->count ? $this->sqlQuery->countQuery() : $this->sqlQuery;
    // Quote arguments so query is able to be run.
    $quoted = [];
    foreach ($sql_query->getArguments() as $key => $value) {
      $quoted[$key] = is_null($value) ? 'NULL' : $this->connection->quote($value);
    }
    // Get sql string prepared for DonorPerfect.
    $sql = $this->prepareSqlForDonorPerfect((string)$sql_query, $quoted);
    // Create the DonorPerfect query.
    $dp_query = new DPQuery('pass', $sql);
    if ($this->count) {
      return (int)$dp_query->execute()->fetchField();
    } else {
      $dp_result = $dp_query->execute()->result;
      $result = [];
      $id_field = $this->entityType->getKey('id');
      foreach ($dp_result as $row) {
        if (is_array($row) && !empty($row[$id_field])) {
          $result[$row[$id_field]] = $row[$id_field];
        }
      }
      return $result;
    }
  }

  /**
   * Implements the magic __toString method.
   */
  public function __toString() {
    // Clone the query so the prepare and compile doesn't get repeated.
    $clone = clone($this);

    $clone->prepare()
      ->compile()
      ->addSort()
      ->finish();

    // Quote arguments so query is able to be run.
    $quoted = [];
    foreach ($clone->sqlQuery->getArguments() as $key => $value) {
      $quoted[$key] = is_null($value) ? 'NULL' : $this->connection->quote($value);
    }

    // Return sql string prepared for DonorPerfect.
    return $this->prepareSqlForDonorPerfect((string)$clone->sqlQuery, $quoted);
  }

  /**
   * Prepare SQL string for DonorPerfect.
   *
   * @param string $sql
   *   The sql string as prepared for MySQL.
   * @param array $arguments
   *   Array of quoted query arguments.
   * 
   * @return string
   */
  public function prepareSqlForDonorPerfect(string $sql, array $arguments=[]) {
    // Clone the query so the prepare and compile doesn't get repeated.
    $clone = clone($this);

    // Replace table name brackets.
    $sql = $clone->connection->prefixTables($sql);
    $sql = $clone->connection->quoteIdentifiers($sql);

    // Replace arguments.
    $sql = strtr($sql, $arguments);

    // Remove double-quotes for DonorPerfect queries.
    $sql = str_replace('"', '', $sql);

    // Remove ESCAPE clauses.
    $sql = preg_replace("/ ESCAPE '.*?'/", '', $sql);

    // Remove line breaks.
    $sql = preg_replace('/[\r\n]/', ' ', $sql);

    return $sql;
  }

}
