<?php

namespace Drupal\donorperfect\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\donorperfect\DPUtility;

/**
 * Base form controller for DonorPerfect entity settings.
 *
 * @ingroup donorperfect
 */
abstract class EntitySettingsFormBase extends ConfigFormBase {

  /**
   * The entity_type_id for which this is the settings form.
   */
  const ENTITY_TYPE_ID = '';

  /**
   * The entity type object.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface
   */
  protected $entityType;

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['donorperfect.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return static::ENTITY_TYPE_ID . '_settings_form';
  }

  /**
   * Gets the entity type object.
   *
   * @return \Drupal\Core\Entity\EntityTypeInterface
   */
  protected function getEntityType() {
    if (empty($this->entityType)) {
      $this->entityType = \Drupal::entityTypeManager()->getDefinition(static::ENTITY_TYPE_ID);
    }
    return $this->entityType;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('donorperfect.settings');
    $trigger = $form_state->getTriggeringElement() ?: ['#name' => ''];
    $entity_type_label = ucfirst($this->getEntityType()->getSingularLabel());
    $base_field = $this->getEntityType()->getKey('id');
    $tables = [
      'base' => $this->getEntityType()->getBaseTable(),
      'udf' => $this->getEntityType()->get('udf_table'),
    ];
    $field_options = [];
    foreach ($tables as $table) {
      $field_options += DPUtility::getFieldsAsOptions($table);
    }
    if (!empty($field_options)) {
      // Alphabetize field list.
      asort($field_options);
      $form[static::ENTITY_TYPE_ID] = [
        '#type' => 'details',
        '#title' => $this->t(':label Entities', [':label' => $entity_type_label]),
        '#tree' => TRUE,
        '#open' => FALSE,
      ];
      $form[static::ENTITY_TYPE_ID]['fields'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Entity Fields'),
        '#description' => $this->t('Select the DonorPerfect fields that should be included in :label entities in Drupal.', [':label' => $entity_type_label]),
        '#description_display' => 'before',
        '#options' => $field_options,
        '#default_value' => $config->get('entity.' . static::ENTITY_TYPE_ID . '.fields') ?: [$base_field],
      ];
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Remove unselected entity fields.
    $values_key = ['entity', static::ENTITY_TYPE_ID, 'fields'];
    $values = $form_state->getValue($values_key, []);
    foreach ($values as $key => $value) {
      if ($key !== $value) unset($values[$key]);
    }
    // Make sure the base field is selected.
    $base_field = $this->getEntityType()->getKey('id');
    $values[$base_field] = $base_field;
    // Set $values back on $form_state.
    $form_state->setValue($values_key, $values);
  }

  /**
   * Perform necessary procedures after form submission is complete.
   *
   * @param array $form
   * @param FormStateInterface $form_state
   */
  public function postSubmitForm(array &$form, FormStateInterface $form_state) {
    // Entity field storage definitions need to be rebuilt.
    /* var \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager */
    $entity_field_manager = \Drupal::service('entity_field.manager');
    $schema_repository = \Drupal::service('entity.last_installed_schema.repository');
    if (!empty($entity_field_manager) && !empty($schema_repository)) {
      $entity_field_manager->clearCachedFieldDefinitions();
      $field_definitions = $entity_field_manager->getBaseFieldDefinitions(static::ENTITY_TYPE_ID);
      $schema_repository->setLastInstalledFieldStorageDefinitions(static::ENTITY_TYPE_ID, $field_definitions);
    }
  }

}
